If[!ValueQ[Global`$MTReloadCheck],
  Global`$MTReloadCheck = True;
];
If[!ValueQ[Global`$MTStartupMessages],
  Global`$MTStartupMessages = True;
];

(* Reload check. *)
If[Global`$MTReloadCheck && MemberQ[$Packages, "MT`"],
  Message[Global`MT::reloaded];
  Abort[];
];
Global`MT::reloaded = "Package MT.m already loaded.";

Global`$MTReloadCheck::usage = "Flag to avoid reloading MT package.";

Global`$MTStartupMessages::usage = "Flag to show messages at loading \
MT package.";

Unprotect["MT`*"];
ClearAll["MT`*", "MT`Private`*"];

BeginPackage["MT`", {"HPL`"}];

If[Global`$MTStartupMessages,
  Print[Style["MT 1.0", Bold],
        " ",
        If[$Notebooks, "\[Copyright]", "(C)"],
        " 2011-2013 by Alexey Pak and Takahiro Ueda, TTP KIT Karlsruhe"];
];

(******************************************************************************)
(* Public functions. *)

HSum::usage = "HSum[{a,b,...},n] is the harmonic sum S_{a,b,...}(n). \
If n is an integer, automatically evaluates to a rational number.";

DReg::usage = "DReg[e,x] is the regularized derivative, i.e. \
the result of differentiation of d e(x)/dx where every singular \
term ln(1-x)^k/(1-x) is replaced with plus-distribution.";

PlusDistribution::usage = "PlusDistribution[i,1-x] is generalized \
function (Log[1-x]^i/(1-x))_+, regularized at x = 1. By \
definition, PlusDistribution[-1,1-x] = delta(1-x).";

MTMellinn::usage = "MTMellinn[e,x,n] performs Mellin transform \
of expression e(x) and returns general n-th term.";

MTInverse::usage = "MTInverse[e,x,n] performs inverse Mellin \
transform of e(n) to a function of x.";

MTNormalize::usage = "MTNormalize[e,n] normalizes the arguments of \
harmonic sums in e so that they have the argument n.";

MTHarmonize::usage = "MTHarmonize[e,n] shifts the arguments of \
harmonic sums in e so that every term in the result depends on \
the same argument (n + i).";

MTProductExpand::usage = "MTProductExpand[e] substitutes products \
of harmonic sums in e with harmonic sums of higher weights.";

MTDRegToPlus::usage = "MTDRegToPlus[e] replaces regularized \
derivatives DReg[...,x] in e with conventional plus-distributions.";

MTPlusToDReg::usage = "MTPlusToDReg[e] replaces \
PlusDistribution[...,1-x] in e with regularized derivatives.";

MTPlusSimplify::usage = "MTPlusSimplify[expr] performs simplifications about \
PlusDistribution.";

Convolution::usage = "Convolution[expr1,expr2,...,x] gives the convolution of \
the given expressions as functions of x.";

CancelMTInverse::usage = "CancelMTInverse[expr] tries to cancel unevaluated \
inverse Mellin transforms in the given expression.";

NConvolution::usage = "NConvolution[expr1,expr2,...,x,value] performs \
the numerical convolution of the given expressions as function of x at \
x = value.";

NEval::usage = "NEval[expr] gives the numerical value of the given expression, \
but does not touch +-distributions.
NEval[expr,x,value] gives the numerical value at x = value.
NEval[expr,n] and NEval[expr,x,value,n] attempt to give a result with \
n-digit precision.";

NMTMellinn::usage = "NMTMellinn[expr,x,value] numerically computes Mellin \
transform of the given expression at n = value.";

$FailedConvs::usage = "$FailedConvs is the list of failed convolutions.";

n$Convolution::usage = "n$Convolution is used for the Mellin parameter in \
unevalauted Mellin transforms.";

(******************************************************************************)

Begin["`Private`"];

(******************************************************************************)
(* File operations. *)

If[$VersionNumber < 7,
  FindFile[name_] := System`Private`FindFile[name];
];

(* Check if the given file is found in $Path. *)
FileFoundQ[name_] := FindFile[name] =!= $Failed;

(******************************************************************************)
(* Compatibilities. *)

If[$VersionNumber < 7,
  DeleteDuplicates[x_List] :=
    Extract[x, Sort[Union[x] /. Dispatch[MapIndexed[Rule, x]]]];
];

(******************************************************************************)
(* Helper functions. *)

(* Check if all indices are integers. *)
IntegerIndicesQ[m_List] := And @@ IntegerQ /@ m /;
  Length[m] >= 1;

IntegerIndicesQ[_] := False;

(* Check if all indices are 0. *)
ZeroIndicesQ[m_List] := And @@ (# === 0 &) /@ m /;
  Length[m] >= 1;

ZeroIndicesQ[_] := False;

(* Check if all indices are 1. *)
OneIndicesQ[m_List] := And @@ (# === 1 &) /@ m /;
  Length[m] >= 1;

OneIndicesQ[_] := False;

(* Check if all indices are not 0. *)
NonZeroIndicesQ[m_List] := And @@ (IntegerQ[#] && # =!= 0 &) /@ m /;
  Length[m] >= 1;

NonZeroIndicesQ[_] := False;

(* Check if all indices are >= 0. *)
NonNegativeIndicesQ[m_List] := And @@ (IntegerQ[#] && # >= 0 &) /@ m /;
  Length[m] >= 1;

NonNegativeIndicesQ[_] := False;

(* Check if the given expression can be expanded by Expand[]. *)
ExpandQ[e_Times] := MemberQ[e, _Plus|Power[_Plus, p_Integer /; p >= 2], {1}];

ExpandQ[Power[_Plus, p_Integer /; p >= 2]] := True;

ExpandQ[_] := False;

(* Expands Log[] with an integer argument if possible. *)
(* Note that PowerExpand[] does not expand Log in Mathematica 6.0. *)
(*
LogExpand[expr_] := expr /. Log[n_] /; IntegerQ[n] && n >= 4 :>
  PowerExpand[Log[n]];
*)
LogExpand[expr_] := expr /. Log[n_] /; IntegerQ[n] && n >= 4 :>
  Plus @@ (#[[2]] Log[#[[1]]] & /@ FactorInteger[n]);

(******************************************************************************)
(* HSum -- memoized for speed *)

mtHSumTerm[k_, i_] := If[k >= 0, i^(-k), (-1)^i * i^k]

HSum[{a___, 0, m1_, b___}, n_] :=
  HSum[{a, Sign[m1] * (Abs[m1] + 1), b}, n] /; IntegerQ[m1] && m1 != 0

HSum[{}, n_] := 1
HSum[l_List, n_Integer] /; NonZeroIndicesQ[l] := mtHSum[l,n] =
  Sum[mtHSumTerm[First[l],i] * HSum[Rest[l],i], {i,n}]

(******************************************************************************)
(* MTMellinn -- generic properties *)

MTMellinn[1, x_, n_] := 1/n;

MTMellinn[c_ * f_., x_, n_] := c * MTMellinn[f, x, n] /; FreeQ[c, x];

MTMellinn[x_^a_. * f_., x_, n_] := MTMellinn[f, x , n + a];

MTMellinn[e_Plus, x_, n_] := MTMellinn[#, x, n]& /@ e;

MTMellinn[e:_Times|_Power, x_, n_] := MTMellinn[e // Expand, x, n] /;
  ExpandQ[e];

(* Known functions. *)

MTMellinn[e:(Log[_]|PolyLog[_, _]|PolyLog[_, _, _])^_. * _., x_, n_] :=
    Block[{convknwnfunstried = True},
  MTMellinn[HPLConvertFromKnownFunctions[e, x], x, n]
] /; !TrueQ[convknwnfunstried];

HPLConvertFromKnownFunctions[expr_, x_] := Module[{ex},
  (* NOTE: we assume 0 < x < 1. *)
  (* Factorize the argument of logarithms if positive. *)
  ex = expr /. {
    Log[z_] :> L[z // Factor, 1]
  } //. {
    L[f_. * x^p_., g_] :> L[f, x^p * g] /; IntegerQ[p],
    L[f_. * (1-x)^p_., g_] :> L[f, (1-x)^p * g] /; IntegerQ[p],
    L[f_. * (1+x)^p_., g_] :> L[f, (1+x)^p * g] /; IntegerQ[p],
    L[f_. * (x-1)^p_., g_] :> L[f * (-1)^p, (1-x)^p * g] /; IntegerQ[p]
  } /. {
    L[a_, b_] :> Log[a * b] /; !TrueQ[Positive[a]],
    L[a_, b_] :> L[a * b] /; TrueQ[Positive[a]]
  } //. {
    L[f_. * x^p_.] :> p * Log[x] + L[f] /; IntegerQ[p],
    L[f_. * (1-x)^p_.] :> p * Log[1-x] + L[f] /; IntegerQ[p],
    L[f_. * (1+x)^p_.] :> p * Log[1+x] + L[f] /; IntegerQ[p]
  } /. {
    L[f_] -> Log[f]
  };
  (* Simplify the arguments of polylogarithms. *)
  ex = ex /. {
    PolyLog[a_, z_] :> PolyLog[a, z // Together // Simplify],
    PolyLog[a_, b_, z_] :> PolyLog[a, b, z // Together // Simplify]
  } /. {
    PolyLog[a_, 2*x/(x-1)] -> PolyLog[a, -2*x/(1-x)],
    PolyLog[a_, b_, 2*x/(x-1)] -> PolyLog[a, b, -2*x/(1-x)]
  };
  (* Convert to HPLs. *)
  (* NOTE: we don't touch PolyLog[k, 0, x], which Mathematica 9 also doesn't
           touch even in a numerical way. *)
  ex /. {
    Log[x] -> HPL[{0}, x],
    Log[1-x] -> - HPL[{1}, x],
    Log[1+x] -> HPL[{-1}, x],
    PolyLog[k_, a:x|-x|x^2|1-x|x/(x-1)|(1-x)/(1+x)|1-x^2|(1+x)/2|(1-x)/2|
                  2*x/(1+x)|-2*x/(1-x)|x/(1+x)] /;
      IntegerQ[k] && k >= 2 -> HPL[{k}, a],
    PolyLog[k_, p_, a:x|-x|x^2|1-x|x/(x-1)|(1-x)/(1+x)|1-x^2|(1+x)/2|(1-x)/2|
                    2*x/(1+x)|-2*x/(1-x)|x/(1+x)] /;
      IntegerQ[k] && k >= 2 && IntegerQ[p] && p >= 2 :>
      HPL[Join[ConstantArray[0, k], ConstantArray[1, p]], a]
  }
];

(* +-distributions. *)

MTMellinn[e:PlusDistribution[k_, 1-x_] * _., x_, n_] :=
  MTMellinn[e // MTPlusSimplify // MTPlusToDReg, x, n] /;
  IntegerQ[k] && k >= -1;

(* Simplification of denominators. *)

MTMellinn[f_. / (x_-1), x_, n_] :=
  - MTMellinn[f / (1-x), x, n]

MTMellinn[e:1 / (1-x_) / (1+x_) * f_., x_, n_] :=
  MTMellinn[Apart[e, x], x, n]

(* Expanding a product of HPLs. *)

MTMellinn[e:HPL[m1_, y_] * HPL[m2_, y_] * f_., x_, n_] :=
  MTMellinn[e // HPLProductExpand, x, n]

MTMellinn[e:HPL[m_, y_]^a_Integer * f_., x_, n_] :=
  MTMellinn[e // HPLProductExpand, x, n] /; a >= 2

(* Transforms the argument of HPLs. *)

MTMellinn[f_. * h:HPL[m_, -x_], x_, n_] :=
  MTMellinn[f * HPLConvertToSimplerArgument[h], x, n] /;
  IntegerIndicesQ[m] && Last[m] =!= 0;

MTMellinn[f_. * h:HPL[m_, x_^2], x_, n_] :=
  MTMellinn[f * HPLConvertToSimplerArgument[h], x, n] /;
  NonNegativeIndicesQ[m];

MTMellinn[f_. * h:HPL[m_, 1-x_], x_, n_] :=
  MTMellinn[f * HPLConvertToSimplerArgument[h], x, n] /;
  NonNegativeIndicesQ[m];

MTMellinn[f_. * h:HPL[m_, x_/(x_-1)], x_, n_] :=
  MTMellinn[f * HPLConvertToSimplerArgument[h], x, n] /;
  NonNegativeIndicesQ[m] && Last[m] =!= 0;

MTMellinn[f_. * h:HPL[m_, (1-x_)/(1+x_)], x_, n_] :=
  MTMellinn[f * HPLConvertToSimplerArgument[h], x, n] /;
  IntegerIndicesQ[m];

MTMellinn[HPL[m_, 1+x_] * f_., x_, n_] :=
  MTMellinn[HPLArgTransform[
            HPLArgTransform[HPL[m, 1+x], x -> 1-x], x -> -x] * f, x, n] /;
  ZeroIndicesQ[m]

MTMellinn[f_. * h:HPL[m_, 1/x_], x_, n_] :=
  MTMellinn[f * HPLArgTransform[h, x -> 1/x], x, n] /;
  NonNegativeIndicesQ[m] && Last[m] === 0 && Count[m, _?Positive] <= 1;

MTMellinn[f_. * h:HPL[m_, 1-x_^2], x_, n_] :=
  MTMellinn[f * (h // HPLArgTransform[#, x -> 1-x]& //
                      HPLConvertToSimplerArgument //
                      Simplify), x, n] /;
  NonNegativeIndicesQ[m];

MTMellinn[f_. * h:HPL[m_, (1+x_)/2], x_, n_] :=
  MTMellinn[f * (h // HPLArgTransform[#, x -> x/(x-1)]& //
                      HPLArgTransform[#, x -> -x]& //
                      HPLArgTransform[#, x -> (1-x)/(1+x)]& //
                      HPLArgTransform[#, x -> -x]& //
                      Simplify), x, n] /;
  NonNegativeIndicesQ[m] && Last[m] =!= 0;

MTMellinn[f_. * h:HPL[m_, (1-x_)/2], x_, n_] :=
  MTMellinn[f * (h // HPLArgTransform[#, x -> x/(x-1)]& //
                      HPLArgTransform[#, x -> -x]& //
                      HPLArgTransform[#, x -> (1-x)/(1+x)]& //
                      Simplify), x, n] /;
  NonNegativeIndicesQ[m] && Last[m] =!= 0;

MTMellinn[f_. * h:HPL[m_, (2*x_)/(1+x_)], x_, n_] :=
  MTMellinn[f * (h // HPLArgTransform[#, x -> 1-x]& //
                      HPLArgTransform[#, x -> (1-x)/(1+x)]& //
                      Simplify), x, n] /;
  NonNegativeIndicesQ[m];

MTMellinn[f_. * h:HPL[m_, -2*x_/(1-x_)], x_, n_] :=
  MTMellinn[f * (h // HPLArgTransform[#, x -> 1-x]& //
                      HPLArgTransform[#, x -> (1-x)/(1+x)]& //
                      HPLArgTransform[#, x -> -x]& //
                      Simplify), x, n] /;
  NonNegativeIndicesQ[m] && Last[m] =!= 0;

MTMellinn[f_. * h:HPL[m_, x_/(1+x_)], x_, n_] :=
  MTMellinn[f * (h // HPLArgTransform[#, x -> x/(x-1)]& //
                      HPLArgTransform[#, x -> -x]& //
                      Simplify), x, n] /;
  NonNegativeIndicesQ[m] && Last[m] =!= 0;

(* With a factor 1/(1+x). *)

MTMellinn[e:HPL[m_, x_] / (1+x_), x_, n_] :=
  MTMellinn[e * DReg[HPL[Prepend[m, -1], x], x]
            / MTDRegToPlus[DReg[HPL[Prepend[m, -1], x], x]] // Cancel, x, n]

MTMellinn[e:1 / (1+x_), x_, n_] :=
  MTMellinn[e * DReg[HPL[{-1}, x], x]
            / MTDRegToPlus[DReg[HPL[{-1}, x], x]] // Cancel, x, n]

MTMellinn[DReg[HPL[m_, x_], x_] / (1+x_), x_, n_] :=
  1/2 * MTMellinn[DReg[HPL[m, x], x], x, n] +
  If[Length[m] == 1,
    1/2 * MTMellinn[1 / (1+x), x, n]
  ,
    1/2 * MTMellinn[HPL[ConstantArray[1, Length[m] - 1], x] / (1+x), x, n]
  ] /;
  OneIndicesQ[m]

(* With a factor 1/(1-x). *)

MTMellinn[e:HPL[m_, x_] / (1-x_), x_, n_] :=
  MTMellinn[e * DReg[HPL[Prepend[m, 1], x], x]
            / MTDRegToPlus[DReg[HPL[Prepend[m, 1], x], x]] // Cancel, x, n] /;
  ZeroIndicesQ[m]

(*
 * We write HPL[{...},x]/(1-x) in terms of DReg[HPL[1,...],x].
 * The cases where all indices of HPL are 1 are excluded.
 *)
MTMellinn[e:HPL[m_List, x_] / (1-x_), x_, n_] := Module[{d, eq, sol, ans, y},
  d = DReg[HPL[Prepend[m, 1], x], x];
  eq = d == MTDRegToPlus[d] /. e -> y;  (* The variable to be solved must be *)
  sol = Solve[eq, y];                   (* a symbol, use y. *)
  If[Length[sol] =!= 1,
    (* The equation should be linear in y, but just in case. *)
    Message[MTMellinn::nodregsol, e];
    Abort[];
  ];
  ans = Collect[y /. First[sol], DReg[_, _]|PlusDistribution[_, _]|HPL[_, _],
                Factor];
  MTMellinn[ans, x, n]
] /; IntegerIndicesQ[m] && Or @@ (# =!= 1 &) /@ m;

MTMellinn::nodregsol = "Can't express `1` via DReg.";

(* DReg(f(x),x) * g(x). *)

MTMellinn[HPL[m_, x_] * DReg[HPL[k_, x_], x_], x_, n_] :=
  MTMellinn[HPL[m, x] * D[HPL[k, x], x], x, n] /;
  ZeroIndicesQ[m] && OneIndicesQ[k]

(* Normalize the third argument. Hope to cancel. *)
MTMellinn[e:1 / (1-x_) * f_., x_, n_ + a_] :=
  MTMellinn[Apart[x^a * e, x], x, n] /;
  IntegerQ[a] && a =!= 0

(******************************************************************************)
(* MTMellinn[DReg[...]]: subtract poles at x = 0, regulate @ x = 1 *)

mtRegTerm[f_HPL,x_] := HPLLogExtract[f] /. x -> 1 /. HPL[{1},1] -> 0

MTMellinn[DReg[1,x_],x_,n_] := 1
MTMellinn[DReg[f_,x_],x_,n_] := mtRegTerm[f,x] -
  (n - 1) * MTMellinn[f,x,n - 1] // Expand

(******************************************************************************)
(* MTInverse *)

MTInverse[e_Plus,x_,n_] := MTInverse[#,x,n]& /@ e
MTInverse[c_,x_,n_] := c * DReg[1,x] /; FreeQ[c,n]
MTInverse[c_*e_,x_,n_] := c * MTInverse[e,x,n] /; FreeQ[c,n]
MTInverse[e:(n_+a1_.)^b1_*(n_+a2_.)^b2_*f_.,x_,n_] :=
  MTInverse[Apart[e, n], x, n] /;
  IntegerQ[a1] && IntegerQ[a2] && IntegerQ[b1] && IntegerQ[b2] &&
  (a1 != 0 || a2 != 0) && b1 <= -1 && b2 <= -1
MTInverse[e:(n_+a_)^b_*f_.,x_,n_] :=
  x^a * MTInverse[Expand[Apart[e /. {n -> n-a}, n]],x,n] /;
  FreeQ[f, HSum]
MTInverse[e:HSum[l_,n_+a_]*f_.,x_,n_] :=
  x^a * MTInverse[Expand[e /. {n -> n-a}],x,n]
MTInverse[MTMellinn[f_,x_,n_],x_,n_] := f

MTInverse[(-1)^(a_ + n_)*f_., x_, n_] := (-1)^a*MTInverse[(-1)^n*f, x, n];

(******************************************************************************)
(* Load tables of Mellin transforms of HPLs and their inverse. *)

MTMellinn::notable = "Table of MTs not found.";
MTInverse::notable = "Table of inverse MTs not found.";

LoadTable[TabName_String] := Module[{weight, fname},
  weight = 0;
  Do[
    fname = TabName <> ToString[i] <> ".mx";
    If[FileFoundQ[fname] &&
     Check[Get[fname], $Failed, {Get::noopen, DumpGet::bgbf}] =!= $Failed,
      weight = i;
      Break[];
    ];
    fname = TabName <> ToString[i] <> ".m";
    If[FileFoundQ[fname] &&
     Check[Get[fname], $Failed, {Get::noopen, DumpGet::bgbf}] =!= $Failed,
      weight = i;
      Break[];
    ];
  , {i, 9, 1, -1}];
  weight
];

TabWeight = LoadTable["MTtab"];

If[TabWeight == 0,
  Message[MTMellinn::notable];
,
  If[Global`$MTStartupMessages,
    Print["Table of MTs loaded up to weight " <> ToString[TabWeight] <> "."];
  ];
];

InvTabWeight = LoadTable["MTitab"];

If[InvTabWeight == 0,
  Message[MTInverse::notable];
,
  If[Global`$MTStartupMessages,
    Print["Table of inverse MTs loaded up to weight " <>
          ToString[InvTabWeight] <> "."];
  ];
];

(******************************************************************************)
(* MTNormalize *)

MTNormalize[expr_, n_] := Module[{ex},
  ex = expr //. {
    HSum[{m1_, mm___}, n + a_] :>
      HSum[{m1, mm}, n] +
        Sum[mtHSumTerm[m1, j] * HSum[{mm}, j], {j, n+1, n+a}] /;
      IntegerQ[m1] && m1 != 0 && IntegerQ[a] && a >= 1,
    HSum[{m1_, mm___}, n + a_] :>
      HSum[{m1, mm}, n] -
        Sum[mtHSumTerm[m1, j] * HSum[{mm}, j], {j, n+a+1, n}] /;
      IntegerQ[m1] && m1 != 0 && IntegerQ[a] && a <= -1
  };
  ex = Apart[ex, n];
  ex = Expand[ex] /. {
    (-1)^(a_ * n) :> (-1)^n /; OddQ[a],
    (-1)^(a_ * n) :> 1 /; EvenQ[a]
  };
  Return[ex];
];

(******************************************************************************)
(* MTHarmonize *)

mtHarmonizeNeg[e_,n_] := PowerExpand[e] /.
  { (-1)^(a_.*n + b_.) /; IntegerQ[a] :> (-1)^If[OddQ[a],n+b,b] } /.
  { (-1)^(n + b_.) * (n + a_.)^c_. :>
      (-1)^(n + a) * (n+a)^c * (-1)^(b - a),
    (-1)^(n + b_.) * HSum[v_,n + a_.]^p_. :>
      (-1)^(n + a) * HSum[v,n + a]^p * (-1)^(b - a) }

mtShift[HSum[l_List,n0_],n_] := Module[{k,ln,i},
  k = First[l]; ln = Rest[l];
  HSum[l,n] -
  Sum[mtHSumTerm[k,i] * HSum[ln,i], {i,n0+1,n}] +
  Sum[mtHSumTerm[k,i] * HSum[ln,i], {i,n+1,n0}]]

$MTHarmonize[expr_, n_] := expr /. {
  e:(n+a_.)^_. * _ :> Expand[Apart[e /. HSum[v_, n+b_.] :>
    mtShift[HSum[v, n+b], n+a] /; a =!= b && IntegerQ[a - b], n]],
  e:(a_.-n)^_. * _ :> Expand[Apart[e /. HSum[v_, n+b_.] :>
    mtShift[HSum[v, n+b], n-a] /; a =!= -b && IntegerQ[a + b], n]],
  c_. * HSum[v_, n+a_.]^p_. * HSum[w_, n+b_.]^q_. :>
    Expand[Apart[c * HSum[v, n+a]^p * mtShift[HSum[w, n+b], n+a]^q, n]] /;
    IntegerQ[p] && p >= 1 && IntegerQ[q] && q >= 1 &&
    a =!= b && IntegerQ[a - b]
};

MTHarmonize[expr_, n_] :=
  FixedPoint[$MTHarmonize[#, n]&, Expand[Apart[expr, n]]] //
  mtHarmonizeNeg[#, n]&;

(******************************************************************************)
(* MTProductExpand *)

mtI[a_,b_] := mtI[(Abs[a] + Abs[b]) * Sign[a] * Sign[b]]
mtN[a___,mtN[],b___] := mtN[a,b]
mtN[a___,k_. * mtN[b__],c___] := k * mtN[a,b,c]
mtN[a___,mtN[b__],c___] := mtN[a,b,c]

mtS[mtN[],mtN[a___]] := mtN[a]
mtS[mtN[a___],mtN[]] := mtN[a]
mtS[mtN[mtI[a1_],as___],mtN[mtI[b1_],bs___]] :=
  Distribute @ mtN[mtI[a1],mtS[mtN[as],mtN[mtI[b1], bs]]] +
  Distribute @ mtN[mtI[b1],mtS[mtN[mtI[a1],as],mtN[bs]]] -
  Distribute @ mtN[mtI[a1,b1],mtS[mtN[as],mtN[bs]]]

mtX[l1_List,l2_List,n_] :=
  mtS[mtN @@ (mtI[#] & /@ l1), mtN @@ (mtI[#] & /@ l2)] /.
  mtN[a__] :> HSum[{a} /. mtI[b_] :> b, n]

MTProductExpand[expr_] := FixedPoint[$MTProductExpand, expr];

$MTProductExpand[expr_] := Expand[Collect[expr, HSum[_, _], coeff]] /. {
  coeff[x_] -> x
} /. {
  HSum[l1_List, n_]^k_ :> mtX[l1, l1, n] * HSum[l1, n]^(k-2) /;
    IntegerQ[k] && k >= 2,
  HSum[l1_List, n_] * HSum[l2_List, n_] :> mtX[l1, l2, n]
};

(******************************************************************************)
(* MTDRegToPlus *)

MTDRegToPlus[expr_] := Module[{ex},
  (* Expand. *)
  ex = expr /. DReg[f_, x_] :> DReg[f // HPLProductExpand // Expand, x] /.
               DReg[e_Plus, x_] :> (DReg[#, x]& /@ e) /.
               DReg[c_ * f_., x_] :> c * DReg[f, x] /; FreeQ[c, x];
  (* Trivial delta functions come from the definition of the Mellin transform
     of DReg[1,x]. *)
  ex = ex /. f_. * DReg[1, x_] :>
             (f /. x -> 1) * PlusDistribution[-1, 1-x];
  (* Regularized derivatives of HPLs multiplied by polynomials. *)
  ex = ex /. {
    DReg[e:p_. * HPL[m_List, x_], x_] :> $DRegToPlus[e, x] /;
      PolynomialQ[p, x] && Length[m] >= 1 && IntegerQ[First[m]],
    DReg[e:p_, x_] :> $DRegToPlus[e, x] /; PolynomialQ[p, x]
  };
  ex
];

$DRegToPlus[expr_, x_] := Module[{ex},
  (* Take a derivative. y = 1-x. *)
  ex = Collect[D[expr, x] /. x -> 1-y // HPLLogExtract,
               {HPL[{1}, 1-y], y}, tt] // Expand;
  ex = ex /. tt[a_] -> a;
  (* Replace 1/(1-x) with +-distributions. *)
  ex = ex /. {
    HPL[{1}, 1-y]^n_. / y -> (-1)^n * PlusDistribution[n, 1-x],
    1 / y -> PlusDistribution[0, 1-x]
  };
  (ex /. y -> 1-x) // MTPlusSimplify // HPLProductExpand
];

(******************************************************************************)
(* MTPlusToDReg *)

MTPlusToDReg[e_] := e /. {
  PlusDistribution[-1, 1-x_] :> DReg[1, x],
  PlusDistribution[n_Integer, 1-x_] :>
    (-1)^n * n! * DReg[HPL[ConstantArray[1, n+1], x], x] /; n >= 0
};

(******************************************************************************)
(* PlusDistribution *)

SetAttributes[PlusDistribution, NHoldAll];

(******************************************************************************)
(* MTPlusSimplify *)

MTPlusSimplify[expr_] :=
  (*
   * Our weapons are
   *   delta(1-x) f(x)
   *     = delta(1-x) f(1),
   *   [ln^k(1-x)/(1-x)]_+ f(x)
   *     = [ln^k(1-x)/(1-x)]_+ f(1) + ln^k(1-x)/(1-x) [f(x) - f(1)].
   *)
  Collect[expr, PlusDistribution[_, 1-x_], Factor] /. {
    PlusDistribution[-1, 1-x_] * f_ :>
      PlusDistribution[-1, 1-x] * (f /. x -> 1),
    PlusDistribution[k_, 1-x_] * f_ /; IntegerQ[k] && k >= 0 :>
      PlusDistribution[k, 1-x] * (f /. x -> 1) +
      (-1)^k * k! * HPL1[k, x] * Factor[(f - (f /. x -> 1)) / (1-x)]
  } /. {
    1/(x_-1) -> - 1/(1-x)
  }

HPL1[0, x_] = 1;
HPL1[k_,x_] := HPL[ConstantArray[1, k], x] /; IntegerQ[k] && k >= 1;

(******************************************************************************)

(*
 * Convolution[expr1, expr2, ..., x] returns the convolution of the given
 * expressions with respect to x.
 *
 * Inverse Mellin transforms can remain in the result of Convolution[].
 * In such cases, n$Convolution is used for the Mellin parameter. Users can
 * sum up results of Convolution[] and try to use CancelMTInverse[].
 *
 * Implementation:
 *   First, the given expressions are expanded and linearity and commutativity
 *   of convolution integrals are used to minimize the number of convolutions
 *   actually to be evaluated. Then they are passed to $Convolution[].
 *
 *   $Convolution[] calculates the convolutions by multiplying Mellin transforms
 *   of the expressions and then subsequently performing the inverse Mellin
 *   transform. The results of $Convolution[] are cached in memory.
 *
 *   With this algorithm Convolution[expr, x] returns expr if it has a Mellin
 *   transform and the inverse Mellin transform is also possible.
 *)

Options[Convolution] = {TimeConstraint -> Infinity,
                        Verbose -> False};

$FailedConvs = {};

Convolution[exprs__, x_, opts : OptionsPattern[]] :=
    Block[{st, n, tmax, convoltried = True}, Module[{e, l},
  (* Check the arguments. *)
  If[Head[x] =!=  Symbol,
    Message[Convolution::ivar, x];
    Return[Convolution[exprs, x, opts]];
  ];
  (* st, n, tmax are used in $Convolution[]. *)
  st = TrueQ[OptionValue[Verbose]] &&
    If[$FrontEnd === Null, Print[##], PrintTemporary[##]] &;
  tmax = OptionValue[TimeConstraint];
  (* Construct an expression such as int[x, HPL[{1},x]]. *)
  e = int @@ (Expand[Apart[#, x]]& /@ {exprs});
  (* Factor out coefficients. *)
  e = e //. {
    int[a___, b_Plus, c___] :> (int[a, #, c]& /@ b),
    int[a___, b_. * c_, d___] /; FreeQ[c, x] && c =!= 1 :>
      LogExpand[c] * int[a, b, d]
  };
  (* Use the internal symbol for x. *)
  e = e /. x -> x$Convolution;
  (* Commutative property of the convolution. *)
  e = e /. int[a___] :> int @@ Sort[{a}];
  (* Perform the convolutions. *)
  e = e /. int[a__] :> $Convolution[a, x$Convolution] /. x$Convolution -> x;
  (* Expand. *)
  e = e // CancelMTInverse // SimplifyX[#, x]&;
  (* Check. *)
  If[!FreeQ[e, MTMellinn],
    l = First /@ Cases[e, MTMellinn[__], {0, Infinity}] // Union;
    If[Length[l] > 0,
      l = { WeightX[#, x], # }& /@ l // Sort;
      If[l[[1]][[1]] === - Infinity,
        Message[Convolution::melx,
                #[[2]]& /@ Select[l, #[[1]] === - Infinity &]];
      ];
      If[TabWeight > 0 && l[[-1]][[1]] > TabWeight,
        Message[Convolution::tab, l[[-1]][[1]], l[[-1]][[2]]];
      ];
    ];
  ];
  If[!FreeQ[e, MTInverse],
    l = First /@ Cases[e, MTInverse[__], {0, Infinity}] // Union;
    l = Select[l, FreeQ[#, MTMellinn]&];  (* Forget about MTMellin. *)
    If[Length[l] > 0,
      l = { WeightN[#, n$Convolution], # }& /@ l // Sort;
      If[l[[1]][[1]] === - Infinity,
        Message[Convolution::imelx,
                #[[2]]& /@ Select[l, #[[1]] === - Infinity &]];
      ];
      If[InvTabWeight > 0 && l[[-1]][[1]] > InvTabWeight,
        Message[Convolution::itab, l[[-1]][[1]], l[[-1]][[2]]];
      ];
    ];
  ];
  (* Return the result. *)
  Return[e];
]] /; !TrueQ[convoltried];

Convolution::ivar = "`1` is not a valid variable.";
Convolution::melx = "Unknown Mellin transform of `1` found.";
Convolution::imelx = "Unknown inverse Mellin transform of `1` found.";
Convolution::tab = "Weight `1` function `2` found. The MT table should be \
expanded.";
Convolution::itab = "Weight `1` function `2` found. The inverse MT table \
should be expanded.";

$Convolution[exprs__, x_] := $Convolution[exprs, x] = Module[{ex},
  ex = TimeConstrained[
    Times@@(MTMellinn[#, x, n]& /@ {exprs})
      // (st["1st expanding..."];
        Expand[#]) &
      // (st["harmonizing..."];
        MTHarmonize[#, n]) &
      // (st["convert products to higher weight..."];
        MTProductExpand[#]) &
      // (st["2nd expanding"];
        Expand[#]) &
      // (st["inverting..."];
        MTInverse[#, x, n]) &
      // (st["handle numeric logarithms..."];
        LogExpand[#])&
      // (st["3rd expanding..."];
        Expand[#]) &
      // (st["establish plus-distributions..."];
        MTPlusSimplify[MTDRegToPlus[#]]) &
      // (st["try to cancel unevaluated MTInverse..."];
        CancelMTInverse[#]) &
      // (st["final expanding..."];
        SimplifyX[#, x] /. n -> n$Convolution) &,
    tmax // Evaluate];
  If[ex === $Aborted,
    AppendTo[$FailedConvs, {exprs}];
    $FailedConvs = DeleteDuplicates[Sort /@ $FailedConvs];
    Message[Convolution::failed, x, {exprs}];
  ];
  ex
];

Convolution::failed = "Unable to do the convolution in `1` of `2`.";

SimplifyX[expr_, x_] := Expand[Apart[expr, x]] /. {
  1/(x-1) -> - 1/(1-x)
};

WeightX[HPL[m_List, x_], x_Symbol] := Total[Max[1, Abs[#]]& /@ m] /;
 And @@ (IntegerQ /@ m);

WeightX[1/(1-x_), x_Symbol] := 0;  (* known as impossible, may cancel *)

WeightX[HPL[m_List, x_] / (1-x_), x_Symbol] := 0 /;
  OneIndicesQ[m];  (* known as impossible, may cancel *)

WeightX[_, _] = - Infinity;  (* for unknown functions *)

WeightN[e_, n_Symbol] := 0 /; FreeQ[e, n];

WeightN[e_Times, n_Symbol] := Total[WeightN[#, n]& /@ List @@ e];

WeightN[(n_ + a_.)^p_., n_Symbol] := - p - 1 /;
  IntegerQ[p] && p <= 0;

WeightN[(-1)^(a_. * n_ + b_.), n_Symbol] := 0 /;
  IntegerQ[a] && IntegerQ[b];

WeightN[HSum[m_List, n_], n_Symbol] := Total[Abs /@ m] /;
  And @@ ((IntegerQ[#] && # != 0)& /@ m);

WeightN[_, _] = - Infinity;  (* for unknown functions *)

(******************************************************************************)

(*
 * CancelMTInverse[expr] tries to cancel out terms containing MTInverse[].
 *)

CancelMTInverse[expr_] := Module[{ex},
  ex = Collect[expr, MTInverse[_, _, _]];
  (* Factor[] may simplify the coefficient and give zero. *)
  ex = ex /. {
    e:f_ * MTInverse[_, _, _] :> Factor[e]
  };
  (* Hopefully unevaluated MTInverse[] cancel if they are combined. *)
  ex = ex /. {
    f_ * MTInverse[a__] :> Expand[f * $MTInverse[a]]
  } /. {
    c_ * $MTInverse[e_, x_, n_] :> $MTInverse[c * e, x, n] /; FreeQ[c, x]
  } /. {
    x_^m_. * $MTInverse[e_, x_, n_] :> $MTInverse[e /. n -> n + m, x, n] /;
      IntegerQ[m]
  } //. {
    $MTInverse[e1_, x_, n_] + $MTInverse[e2_, x_, n_] ->
      $MTInverse[e1 + e2, x, n]
  } /. {
    $MTInverse[e_, x_, n_] :> MTInverse[MTNormalize[e, n], x, n]
  };
  Return[ex];
];

(******************************************************************************)

(*
 * NConvolution[expr1, expr2, ..., x, xvalue] returns the convolution of
 * the given expressions with respect to x numerically evaluated at x = xvalue.
 *
 * Implementation:
 *   NConvolution[] is actually a wrapper function of NConvolutionImpl[] and
 *   processes the options. It also checks if there are unevaluated
 *   convolutions at the end.
 *
 *   NConvolutionImpl[] tries to minimize the number of convolutions actually
 *   to be evaluated. As a result, the convolution is expressed as a linear
 *   combination of $NConvolution[]. The main problem comes from the fact that
 *   in the analytical convolution possible logarithmic divergences are expected
 *   to be expressed as powers of HPL[{1}, 1] and in this sense each integral
 *   are regularised even if they are split by the distributivity and separately
 *   divergent, but in the numerical convolution we should not split an integral
 *   to a sum of divergent integrals. Therefore NConvolutionImpl[] splits
 *   convolutions only with respect to +-distributions and objects explicitly
 *   given by Constants option.
 *
 *   In $NConvolution[], firstly delta functions are removed. If $NConvolution[]
 *   has only three arguments, i.e., only one expression is given as
 *   $NConvolution[expr, x, xvalue], no integration is needed.
 *   If $NConvolution[] is actually a convolution of two or more expressions and
 *   all of them are regular functions, the convolution integral is
 *   straightforward. If the convolution contains +-distributions, we use
 *   a regularisation described in Sect. 5 of arXiV:hep-ph/0501130 by
 *   C. Anastasiou, K. Melnikov and F. Petriello. The regularised expression is
 *   passed to $NConvolutionRegularized[].
 *
 *   Currently the following types of convolutions are implemented:
 *     F1(x) * F2(x)
 *     F1(x) * F2(x) * F3(x)
 *     F1(x) * F2(x) * F3(x) * F4(x)
 *     D1(x) * F2(x)
 *     D1(x) * F2(x) * F3(x)
 *     D1(x) * D2(x)
 *     D1(x) * D2(x) * F3(x)
 *     D1(x) * D2(x) * D3(x)
 *   where F's are regular functions and D's are +-distributions.
 *)

Options[NConvolution] = {
  AccuracyGoal     -> Infinity,
  Constants        -> {},
  MaxPoints        -> Automatic,
  MaxRecursion     -> Automatic,
  Method           -> Automatic,
  MinRecursion     -> 0,
  PrecisionGoal    -> Automatic,
  Tolerance        -> 10^-6,
  WorkingPrecision -> MachinePrecision
};

NConvolution[exprs__, x_, xvalue_, opts:OptionsPattern[]] :=
    Block[{consts, neval, nintegrate, eps, wp, nintopts, nconvoltried = True},
    Module[{e},
  (* Check the arguments. *)
  If[Head[x] =!=  Symbol,
    Message[NConvolution::ivar, x];
    Return[NConvolution[exprs, x, xvalue, opts]];
  ];
  If[!NumericQ[xvalue],
    Message[NConvolution::nonum];
    Return[NConvolution[exprs, x, xvalue, opts]];
  ];
  (* Process options. *)
  consts = OptionValue[Constants];
  If[!ListQ[consts],
    consts = { consts }
  ];
  eps = OptionValue[Tolerance];
  wp = OptionValue[WorkingPrecision];
  nintopts = {
    AccuracyGoal     -> OptionValue[AccuracyGoal],
    MaxPoints        -> OptionValue[MaxPoints],
    MaxRecursion     -> OptionValue[MaxRecursion],
    Method           -> OptionValue[Method],
    MinRecursion     -> OptionValue[MinRecursion],
    PrecisionGoal    -> OptionValue[PrecisionGoal],
    WorkingPrecision -> wp
  };
  nintopts = Sequence @@
             Select[nintopts, (OptionValue[NIntegrate, #[[1]]] =!= #[[2]])&];
  neval[a__] := NEval[a, wp];
  nintegrate[a__] := NInteg[a, nintopts];
  (* Perform the convolution. *)
  e = NConvolutionImpl[exprs, x, xvalue];
  If[!FreeQ[e, $NConvolution],
    Message[NConvolution::notimpl];
    e = e /. $NConvolution[a__] :> NConvolution[a, opts];
  ];
  Return[e];
]] /; !TrueQ[nconvoltried];

NConvolution::ivar = "`1` is not a valid variable.";
NConvolution::nonum = "The last argument of NConvolution is supposed to be a \
number.";
NConvolution::notimpl = "Not implemented convolution.";

NConvolutionImpl[exprs__, x_, xvalue_] := Module[{ex},
  ex = int[exprs];
  (*
   * Classify terms in the integrand by +-distributions and split the integral
   * by them.
   *)
  ex = (# // MTDRegToPlus // MTPlusSimplify)& /@ ex;
  ex = Expand[Collect[#, Append[consts, PlusDistribution[_, _]], coeff]]& /@
         ex //. {
    int[e1___, e_Plus, e2___] :> (int[e1, #, e2]& /@ e)
  } /. {
    coeff[a_] -> a
  };
  (* Simplification by extracting overall factors. *)
  ex = ex //. {
    int[e1___, c_ * e_., e2___] /; FreeQ[c, x] :> c * int[e1, e, e2]
  };
  (* Sorting. *)
  ex = ex /. {
    int[e___] :>  int @@ Sort[{e}, NConvolOrderedQ]
  };
  (* Numerical evaluation. *)
  ex = Collect[ex, int[___]] /. int[e___] :> $NConvolution[e, x, xvalue];
  ex // Expand // neval
];

NConvolOrderedQ[e1_, e2_] := Module[{l1, l2, k1, k2},
  {l1, l2} = Cases[#, PlusDistribution[k_, _], {0, Infinity}]& /@ {e1, e2};
  If[Length[l1] > Length[l2], Return[True]];
  If[Length[l1] < Length[l2], Return[False]];
  If[Length[l1] == 0, Return[True]];
  {k1, k2} = {l1, l2} /. PlusDistribution[k_, _] -> k;
  If[Min[k1] > Min[k2], Return[False]];
  True
];

SetAttributes[$NConvolution, NHoldAll];

(* Evaluation of one function. *)

$NConvolution[e_, x_, xvalue_] := Eval[e, x, xvalue] /;
  FreeQ[e, PlusDistribution];

$NConvolution[e:PlusDistribution[k_, 1-x_], x_, xvalue_] :=
  Eval[e, x, xvalue] /; IntegerQ[k] && k >= -1;

(* Removing a delta function. *)

$NConvolution[e1___, PlusDistribution[-1, 1-x_], e2___, x_, xvalue_] :=
  $NConvolution[e1, e2, x, xvalue] /; Length[{e1, e2}] >= 1;

(* Convolutions for regular functions. *)

$NConvolution[e1_, e2_, x_, xvalue_] :=
  nintegrate[1 / y (e1 /. x -> y) (e2 /. x -> xvalue / y), {y, xvalue, 1}] /;
  FreeQ[{e1, e2}, PlusDistribution];

$NConvolution[e1_, e2_, e3_, x_, xvalue_] :=
  nintegrate[1 / y1 / y2
             (e1 /. x -> y1) (e2 /. x -> y2) (e3 /. x -> xvalue / y1 / y2 ),
             {y1, xvalue, 1}, {y2, xvalue / y1, 1}] /;
  FreeQ[{e1, e2, e3}, PlusDistribution];

$NConvolution[e1_, e2_, e3_, e4_, x_, xvalue_] :=
  nintegrate[1 / y1 / y2 / y3
             (e1 /. x -> y1) (e2 /. x -> y2) (e3 /. x -> y3)
             (e4 /. x -> xvalue / y1 / y2 / y3),
             {y1, xvalue, 1}, {y2, xvalue / y1, 1},
             {y3, xvalue / y1 / y2, 1}] /;
  FreeQ[{e1, e2, e3, e4}, PlusDistribution];

(* Convolutions with one +-distribution. *)

(* Hard coded without $NConvolutionRegularized[] for speed improvements. *)
$NConvolution[PlusDistribution[k_, 1-x_], e2_, x_, xvalue_] :=
  nintegrate[Log[1-y]^k / (1-y) * (
        1 / y * (e2 /. x -> xvalue / y)
        - 1 * (e2 /. x -> xvalue)
      ), {y, xvalue, 1}] - (e2 /. x -> xvalue) *
    nintegrate[Log[1-y]^k / (1-y), {y, 0, xvalue}] /;
  FreeQ[e2, PlusDistribution] && IntegerQ[k] && k >= 0;

(*
$NConvolution[PlusDistribution[k_, 1-x_], e2_, x_, xvalue_] :=
  $NConvolutionRegularized[
    (1-x)^(a*ep) * int[y1] / (x+(1-x)*y1) * (1-y1)^(-1+a*ep) *
        (e2 /. x -> x/(x+(1-x)*y1)) -
      e2 / (a*ep),
    {a}, {k}, x, xvalue, int, ep] /;
  FreeQ[e2, PlusDistribution] && IntegerQ[k] && k >= 0;
*)

$NConvolution[PlusDistribution[k_, 1-x_], e2_, e3_, x_, xvalue_] :=
  $NConvolutionRegularized[
    (1-x)^(1+a*ep) * int[y1] * y1 / (x+(1-x)*y1) * int[y2] / (x+(1-x)*y1*y2) *
        (1-y1)^(-1+a*ep) *
        (e2 /. x -> (x+(1-x)*y1*y2)/(x+(1-x)*y1)) *
        (e3 /. x -> x/(x+(1-x)*y1*y2)) -
      1/(a*ep) * (1-x) * int[y2] / (x+(1-x)*y2) *
        (e2 /. x -> x+(1-x)*y2) *
        (e3 /. x -> x/(x+(1-x)*y2)),
    {a}, {k}, x, xvalue, int, ep] /;
  FreeQ[{e2, e3}, PlusDistribution] && IntegerQ[k] && k >= 0;

(* Convolutions with two +-distributions. *)

$NConvolution[PlusDistribution[k1_,1-x_],
              PlusDistribution[k2_,1-x_], x_, xvalue_] :=
  $NConvolutionRegularized[
    (1-x)^(-1+(a+b)*ep) * int[y1] * y1^(-1+b*ep) * (1-y1)^(-1+a*ep) *
        (x+(1-x)*y1)^(-b*ep) -
      (1-x)^(-1+a*ep) / (b*ep) -
      (1-x)^(-1+b*ep) / (a*ep) +
      PlusDistribution[-1, 1-x] / (a*b*ep^2),
    {a, b}, {k1, k2}, x, xvalue, int, ep] /;
  IntegerQ[k1] && k1 >= 0 && IntegerQ[k2] && k2 >= 0;

$NConvolution[PlusDistribution[k1_,1-x_],
              PlusDistribution[k2_,1-x_], e3_, x_, xvalue_] :=
  $NConvolutionRegularized[
    (1-x)^((a+b)*ep) * int[y1] * int[y2] * (1-y1)^(-1+a*ep) * y1^(b*ep) *
        (x+(1-x)*y1)^(-b*ep) * (1-y2)^(-1+b*ep) / (x+(1-x)*y1*y2) *
        (e3 /. x -> x/(x+(1-x)*y1*y2)) -
      1/(a*ep) * (1-x)^(b*ep) * int[y2] * (1-y2)^(-1+b*ep) / (x+(1-x)*y2) *
        (e3 /. x -> x/(x+(1-x)*y2)) -
      1/(b*ep) * (1-x)^(a*ep) * int[y1] * (1-y1)^(-1+a*ep) / (x+(1-x)*y1) *
        (e3 /. x -> x/(x+(1-x)*y1)) +
      1/(a*b*ep^2) * e3,
    {a, b}, {k1, k2}, x, xvalue, int, ep] /;
  FreeQ[e3, PlusDistribution] && IntegerQ[k1] && k1 >= 0 &&
  IntegerQ[k2] && k2 >= 0;

(* Convolutions with three +-distributions. *)

$NConvolution[PlusDistribution[k1_,1-x_], PlusDistribution[k2_,1-x_],
              PlusDistribution[k3_,1-x_], x_, xvalue_] :=
  $NConvolutionRegularized[
    (1-x)^(-1+(a+b+c)*ep) * int[y1] * int[y2] * y1^(-1+(b+c)*ep) *
        (1-y1)^(-1+a*ep) * y2^(-1+c*ep) * (1-y2)^(-1+b*ep) *
        (x+(1-x)*y1)^(-b*ep) * (x+(1-x)*y1*y2)^(-c*ep) -
    1/(a*ep) * (1-x)^(-1+(b+c)*ep) * int[y2] * y2^(-1+c*ep) * (1-y2)^(-1+b*ep) *
        (x+(1-x)*y2)^(-c*ep) -
    1/(b*ep) * (1-x)^(-1+(a+c)*ep) * int[y1] * y1^(-1+c*ep) * (1-y1)^(-1+a*ep) *
        (x+(1-x)*y1)^(-c*ep) -
    1/(c*ep) * (1-x)^(-1+(a+b)*ep) * int[y1] * y1^(-1+b*ep) * (1-y1)^(-1+a*ep) *
        (x+(1-x)*y1)^(-b*ep) +
    1/(a*b*ep^2) * (1-x)^(-1+c*ep) +
    1/(a*c*ep^2) * (1-x)^(-1+b*ep) +
    1/(b*c*ep^2) * (1-x)^(-1+a*ep) -
    1/(a*b*c*ep^3) * PlusDistribution[-1, 1-x],
    {a, b, c}, {k1, k2, k3}, x, xvalue, int, ep] /;
  IntegerQ[k1] && k1 >= 0 && IntegerQ[k2] && k2 >= 0 && IntegerQ[k3] && k3 >= 0;

(* Convolution integral regularized as hep-ph/0501130 Sect. 5. *)

$NConvolutionRegularized[expr_, a_List, m_List, x_, xvalue_, int_, ep_] :=
    Module[{ex},
  ex = expr / ep^Total[m];
  (* Expansion by +-distribution. *)
  ex = ExpandSingularIntegrals[ex, int, ep];
  ex = ExpandByPlusDistributions[ex, x, ep, - Exponent[ex, ep, Min]];
  (* Take the limit. *)
  Do[
    ex = D[ex, {a[[i]], m[[i]]}] /. a[[i]] -> 1;
  , {i, 1, Length[a]}];
  ex = Series[ex, {ep, 0, 0}] // Normal // Expand;
  (* Collect integrals. *)
  ex = ex //. {
    int[y1__] * int[y2__] -> int[y1, y2]
  };
  ex = Collect[ex, {ep, int[__]}, MTPlusSimplify];
  ex = Collect[ex, {ep, int[__], PlusDistribution[_, 1-x]},
               coeff[# // Simplify]&] /. coeff[0] -> 0 // Expand;
  (* Numerical integration. *)
  ex = Eval[ex, x, xvalue] /. {
    (* All pole terms are expected to cancel. *)
    int[yy__] * coeff[e_] * ep^p_ :>
      chop[nintegrate[e, Sequence @@ ({#, 0, 1} & /@ {yy}),
           AccuracyGoal -> -Log10[eps]]] * ep^p /;
        IntegerQ[p] && p <= -1,
    coeff[e] * ep^p_ :>
      chop[N[e, wp]] * ep^p /; IntegerQ[p] && p <= -1,
    int[yy__] * coeff[e_] :>
      nintegrate[e, Sequence @@ ({#, 0, 1} & /@ {yy})],
    coeff[e_] -> e
  } /. {
    (* Drop terms which are almost zero as expected. *)
    chop[y_] :> Chop[y, eps]
  };
  (* Check poles. *)
  If[!FreeQ[ex, ep],
    Message[NConvolution::pole];
  ];
  (* Return. *)
  Return[ex];
];

NConvolution::pole = "Uncancelled regulators appear in the result.";

SplitSingularIntegrals[expr_, int_, ep_] :=
  Expand[expr] /. {
    int[y_] * y_^(-1+a_.*ep) * (1-y_)^(-1+b_.*ep) * e_. :>
      1/2^(a*ep) * int[y] * y^(-1+a*ep) * (1-y/2)^(-1+b*ep) *
        (e /. y -> y/2) +
      1/2^(b*ep) * int[y] * (1-y/2)^(-1+a*ep) * y^(-1+b*ep) *
        (e /. y -> 1 - y/2)
  };

ExpandSingularIntegrals[expr_, int_, ep_] := Module[{f},
  f[ex_] := Expand[ex] /. {
    int[y_] * y_^(-1+a_.*ep) * e_. :>
      (e /. y -> 0) / (a*ep) +
      int[y] * y^(a*ep) * (e - (e /. y -> 0)) * den[y],
    int[y_] * (1-y_)^(-1+a_.*ep) * e_. :>
      (e /. y -> 1) / (a*ep) +
      int[y] * (1-y)^(a*ep) * (e - (e /. y -> 1)) * den[1-y]
  };
  Collect[expr, ep, FixedPoint[f, #]&] /. den[y_] -> 1/y
];

ExpandByPlusDistributions[expr_, x_, ep_, n_] :=
  Expand[expr] /. {
    (1-x)^(-1+a_.*ep) :>
      PlusDistribution[-1, 1-x] / (a*ep) +
        Sum[(a*ep)^k / k! * PlusDistribution[k, 1-x], {k, 0, n}]
  };

(*
 * Eval[expr, x, xvalue] is the same as (expr /. x -> xvalue) except that it
 * does not touch +-distributions.
 *)
Eval[expr_, x_Symbol, xvalue_] := Module[{ex, y},
  ex = MTDRegToPlus[expr] /. {
    d_PlusDistribution :> (d /. x -> y)
  };
  ex = ex /. x -> xvalue;
  ex = ex /. {
    y -> x
  };
  Return[ex];
];

(*
 * Customized NIntegrate[] routine.
 *)
NInteg[f_, a__, opts:OptionsPattern[]] := Module[{ex},
  If[TrueQ[f == 0],
    (* Avoid NIntegrate::izero. *)
    Return[0];
  ];
  ex = PrepareNumericIntegrand[f, a, wp];
  If[$NIntegrateVerbose,
    Print[{ex, a} /. {y -> Global`y, y1 -> Global`y1, y2 -> Global`y2}];
  ];
  ex = NIntegrate @@ { ex, a, opts };
  If[$NIntegrateVerbose,
    Print[ex];
  ];
  ex
];

(*
 * PrepareNumericIntegrand[expr, n] makes the given expression suitable for
 * an integrand of NIntegrate[] with n-digit.
 *
 * We translate HPLs to known functions as much as possible. This is very
 * important for speed.
 *
 * NumericQ[HPL[{1},1/2]] is False (with HPL 2.0 (03-NOV-2011)) and then
 * NIntegrate[] complains non-numerical values for all sampling points.
 * Explicit N[] is needed.
 *
 * XXX: HPLConvertToKnownFunctions[HPL[{1}, 1]] gives a term containing
 *      HPL`Private`DoConvert (with HPL 2.0 (03-NOV-2011)), although it is
 *      divergent.
 *)
PrepareNumericIntegrand[expr_, a:(_List)..., n_] := Module[{ex},
  ex = expr // HPLLogExtract // HPLConvertToKnownFunctions;
  ex = ex /. {
    HPL[m_, x_]     :> HPL[m, Simplify[x]],
    PolyLog[k_, x_] :> PolyLog[k, Simplify[x]],
    Log[x_]         :> Log[Simplify[x]]
  };
  (* Integration by parts for 1/(1-x). *)
  If[TrueQ[$UseIBP],
    If[FreeQ[ex, HPL], (* Limit does not work with HPL. *)
      Module[{aa, x, ee, e1, e2, lim0, lim1},
        aa = {a};
        Do[
          If[aa[[i]][[2]] === 0 && aa[[i]][[3]] === 1,
            x = aa[[i]][[1]];
            ee = Apart[ex, x] + dum1 + dum2 /. 1 / (x-1) -> - 1 / (1-x);
            e1 = Select[ee, Head[#] === Times && MemberQ[#, 1 / (1-x)]&];
            e2 = ee - e1;
            e1 = (((1-x)#)& /@ (e1 + dum1 + dum2)) /. {dum1 -> 0, dum2 -> 0};
            lim1 = Limit[-Log[1-x] * e1, x -> 1, Direction -> 1];
            lim0 = Limit[-Log[1-x] * e1, x -> 0, Direction -> 0];
            If[FiniteQ[lim1] && FiniteQ[lim0],
              ex = lim1 - lim0 + Log[1-x] * D[e1, x] +
                   e2 /. { dum1 -> 0, dum2 -> 0};
            ];
          ],
        {i, Length[aa]}];
      ];
    ];
  ];
  ex = Collect[ex, {HPL[_, x_] /; !NumericQ[x], PolyLog[_, x_] /; !NumericQ,
                    Log[x_] /; !NumericQ[x]}, Factor];
  ex = ex /. {
    h:HPL[{_Integer..}, _?NumericQ] :> N[h, n + 10]
  };
  ex
];

FiniteQ[DirectedInfinity[_]] := False;
FiniteQ[DirectedInfinity[]] := False;
FiniteQ[Indeterminate] := False;
FiniteQ[_] := True;

(* Fixes for HPL 2.0 (03-NOV-2011). *)
NumericQ[HPLs6]  = True;
NumericQ[HPLs7a] = True;
NumericQ[HPLs7b] = True;
NumericQ[HPLs8a] = True;
NumericQ[HPLs8b] = True;
NumericQ[HPLs8c] = True;
NumericQ[HPLs8d] = True;

(******************************************************************************)

(*
 * NEval[expr] is same as N[expr] but does not touch HPL[{1}, 1].
 *
 * NEval[expr, x, xvalue] is same as N[expr /. x -> xvalue] but does not touch
 * +-distributions as well as HPL[{1}, 1].
 *
 * NEval[expr, n] and NEval[expr, x, xvalue, n] attempt to give a result with
 * n-digit precision.
 *)

NEval[expr_, n_:MachinePrecision] := Module[{y},
  NEval[expr, y, 1, n]
];

NEval[expr_, x_Symbol, xvalue_?NumericQ, n_:MachinePrecision] :=
    Module[{ex, y, L},
  ex = MTDRegToPlus[expr] /. {
    d_PlusDistribution :> (d /. x -> y),
    HPL[{1}, 1] -> L
  };
  ex = ex /. x -> xvalue;
  ex = If[n === MachinePrecision, N[ex], N[ex, n]];
  ex = ex /. {
    y -> x,
    L -> HPL[{1}, 1]
  };
  Return[ex];
];

(******************************************************************************)

(*
 * NMTMellinn[expr, x, nvalue] numerically computes the Mellin transform of
 * the given expression with respect to x at n = nvalue.
 *
 * Implementation:
 *   After applying MTDRegToPlus[], we split the integral by the specified
 *   Constants and +-distributions. Each integral is straightforwardly
 *   evaluated.
 *)

Options[NMTMellinn] = {
  AccuracyGoal     -> Infinity,
  Constants        -> {},
  MaxPoints        -> Automatic,
  MaxRecursion     -> Automatic,
  Method           -> Automatic,
  MinRecursion     -> 0,
  PrecisionGoal    -> Automatic,
  WorkingPrecision -> MachinePrecision
};

NMTMellinn[expr_, x_Symbol, nvalue_?NumericQ, opts:OptionsPattern[]] :=
    Block[{nmellintried = True, wp, nintopts, neval, nintegrate},
    Module[{ex, const},
  (* Process options. *)
  consts = OptionValue[Constants];
  If[!ListQ[consts],
    consts = { consts }
  ];
  wp = OptionValue[WorkingPrecision];
  nintopts = {
    AccuracyGoal     -> OptionValue[AccuracyGoal],
    MaxPoints        -> OptionValue[MaxPoints],
    MaxRecursion     -> OptionValue[MaxRecursion],
    Method           -> OptionValue[Method],
    MinRecursion     -> OptionValue[MinRecursion],
    PrecisionGoal    -> OptionValue[PrecisionGoal],
    WorkingPrecision -> wp
  };
  nintopts = Sequence @@
             Select[nintopts, (OptionValue[NIntegrate, #[[1]]] =!= #[[2]])&];
  neval[a__] := NEval[a, wp];
  nintegrate[a__] := NInteg[a, nintopts];
  (*
   * Classify terms in the integrand by +-distributions and split the integral
   * by them.
   *)
  ex = int[Expand[Collect[MTDRegToPlus[expr],
                   Append[consts, PlusDistribution[_, _]], coeff]]] /. {
    int[e_Plus] :> (int /@ e)
  } /. {
    coeff[e_] -> e
  };
  (* Simplification by extracting overall factors. *)
  ex = ex /. {
    int[c_ * e_.] :> c * int[e] /; FreeQ[c, x]
  };
  (* Numerical evaluation. *)
  ex = Collect[ex, int[_]] /. int[e_] :> $NMTMellinn[e, x, nvalue];
  ex = ex // Expand // neval;
  If[!FreeQ[ex, $NMTMellinn],
    Message[NMTMellinn::notimpl];
    ex = ex /. $NMTMellinn[a__] :> NMTMellinn[a, opts];
  ];
  Return[ex];
]] /; !TrueQ[nmellintried];

NMTMellinn::notimpl = "Not implemented integral.";

SetAttributes[$NMTMellinn, NHoldAll];

$NMTMellinn[e_, x_, nvalue_] :=
  nintegrate[x^(nvalue - 1) * e, {x, 0, 1}] /;
  FreeQ[e, PlusDistribution];

$NMTMellinn[e_. * PlusDistribution[-1, 1-x_], x_, nvalue_] :=
  Eval[e, x, 1] /;
  FreeQ[e, PlusDistribution];

$NMTMellinn[e_. * PlusDistribution[k_, 1-x_], x_, nvalue_] :=
  nintegrate[(x^(nvalue - 1) * e - (e /. x -> 1)) * Log[1-x]^k / (1-x),
             {x, 0, 1}] /;
  IntegerQ[k] && k >= 0 && FreeQ[e, PlusDistribution];

(******************************************************************************)
(* Fancy output *)

MakeIndicesBoxes[m_List, form_] :=
  RowBox[Riffle[MakeBoxes[#, form]& /@ m, ","]];

Unprotect[HPL, MZV];

HPL /: MakeBoxes[HPL[m_List, x_], TraditionalForm] :=
  InterpretationBox @@ {
    RowBox[{SubscriptBox["H", MakeIndicesBoxes[m, TraditionalForm]], "(",
            MakeBoxes[x, TraditionalForm], ")"}],
    HPL[m, x]};

MZV /: MakeBoxes[MZV[m_List], TraditionalForm] :=
  InterpretationBox @@ {
    RowBox[{SubscriptBox["\[Zeta]", MakeIndicesBoxes[m, TraditionalForm]]}],
    MZV[m]};

HPLs6 /: MakeBoxes[HPLs6, TraditionalForm] :=
  InterpretationBox @@ {
    MakeBoxes[HSum[{-5, -1}, Infinity], TraditionalForm],
    HPLs6};

HPLs7a /: MakeBoxes[HPLs7a, TraditionalForm] :=
  InterpretationBox @@ {
    MakeBoxes[HSum[{-5, 1, 1}, Infinity], TraditionalForm],
    HPLs7a};

HPLs7b /: MakeBoxes[HPLs7b, TraditionalForm] :=
  InterpretationBox @@ {
    MakeBoxes[HSum[{5, -1, -1}, Infinity], TraditionalForm],
    HPLs7b};

HPLs8a /: MakeBoxes[HPLs8a, TraditionalForm] :=
  InterpretationBox @@ {
    MakeBoxes[HSum[{5, 3}, Infinity], TraditionalForm],
    HPLs8a};

HPLs8b /: MakeBoxes[HPLs8b, TraditionalForm] :=
  InterpretationBox @@ {
    MakeBoxes[HSum[{-7, -1}, Infinity], TraditionalForm],
    HPLs8b};

HPLs8c /: MakeBoxes[HPLs8c, TraditionalForm] :=
  InterpretationBox @@ {
    MakeBoxes[HSum[{-5, -1, -1, -1}, Infinity], TraditionalForm],
    HPLs8c};

HPLs8d /: MakeBoxes[HPLs8d, TraditionalForm] :=
  InterpretationBox @@ {
    MakeBoxes[HSum[{-5, -1, 1, 1}, Infinity], TraditionalForm],
    HPLs8d};

HSum /: MakeBoxes[HSum[m_List, n_], TraditionalForm] :=
  InterpretationBox @@ {
    RowBox[{SubscriptBox["S", MakeIndicesBoxes[m, TraditionalForm]], "(",
            MakeBoxes[n, TraditionalForm], ")"}],
    HSum[m, n]};

DReg /: MakeBoxes[DReg[e_Plus|e_Times, x_], TraditionalForm] :=
  InterpretationBox @@ {
    RowBox[{SubscriptBox[OverscriptBox["\[PartialD]", "^"],
                         MakeBoxes[x, TraditionalForm]],
            "(", MakeBoxes[e, TraditionalForm], ")"}],
    DReg[e, x]};

DReg /: MakeBoxes[DReg[e_, x_], TraditionalForm] :=
  InterpretationBox @@ {
    RowBox[{SubscriptBox[OverscriptBox["\[PartialD]", "^"],
                         MakeBoxes[x, TraditionalForm]],
            MakeBoxes[e, TraditionalForm]}],
    DReg[e, x]};

PlusDistribution /: MakeBoxes[
    PlusDistribution[-1, x:_Symbol|1-_Symbol], TraditionalForm] :=
  InterpretationBox @@ {
    RowBox[{"\[Delta]", "(", MakeBoxes[x, TraditionalForm], ")"}],
    PlusDistribution[-1, x]};

PlusDistribution /: MakeBoxes[
    PlusDistribution[0, x:_Symbol|1-_Symbol], TraditionalForm] :=
  InterpretationBox @@ {
    SubscriptBox[RowBox[{"[", MakeBoxes[1/x, TraditionalForm], "]"}], "+"],
    PlusDistribution[0, x]};

PlusDistribution /: MakeBoxes[
    PlusDistribution[1, x:_Symbol|1-_Symbol], TraditionalForm] :=
  InterpretationBox @@ {
    SubscriptBox[RowBox[{"[", MakeBoxes[(Log[x])/x, TraditionalForm],
                         "]"}], "+"],
    PlusDistribution[1, x]};

PlusDistribution /: MakeBoxes[
    PlusDistribution[n_?(IntegerQ[#] && # >= 2&),
                     x:_Symbol|1-_Symbol], TraditionalForm] :=
  InterpretationBox @@ {
    SubscriptBox[RowBox[{"[", MakeBoxes[(Log[x]^n)/(x), TraditionalForm],
                         "]"}], "+"],
    PlusDistribution[n, x]};

Protect[HPL, MZV];

(******************************************************************************)

(* Protect symbols in the package. *)
Scan[SetAttributes[#, {Protected, ReadProtected}]&,
     Select[Symbol /@ Names["MT`*"], Head[#] === Symbol &]];

If[Global`$MTStartupMessages,
  Print["See ?MT`* for a list of functions."];
];

End[]
EndPackage[]

(* vim: set ft=mma et ts=8 sts=2 sw=2: *)
