(*
 * MakeTab.m
 *
 * Makes the table for Mellin transform of HPLs for MT.m using
 * the FORM package harmpol.h.
 *
 * Needs the HPL package, FORM.m, summer.h and harmpol.h.
 *
 * Usage:
 *   math < MakeTab.m
 * or
 *   math -run 'MaxWeight=6;FORMBin="form";<<MakeTab`;Quit[]'
 *)

Needs["HPL`"];
Needs["FORM`"];

(* The maximum weight of the tables. *)
If[!ValueQ[MaxWeight],
  MaxWeight = 6;
];
(* The path to the FORM executable. *)
If[!ValueQ[FORMBin],
  FORMBin = "form";
];
(* Flag to create also the table for regularised derivatives. *)
If[!ValueQ[MakeDRegTab],
  MakeDRegTab = True;
];

TabFileName     = "MTtab"  <> ToString[MaxWeight] <> ".m";
DRegTabFileName = "MTdtab" <> ToString[MaxWeight] <> ".m";
InvTabFileName  = "MTitab" <> ToString[MaxWeight] <> ".m";

(******************************************************************************)

SetOptions[FORM, Print -> False, FORMBIN -> FORMBin];

HMellin[expr_, x_, n_] := Module[{ex, out},
  ex = expr //. {
    HPL[{m__}, y_] -> H[R[m], y],
    x -> tmpxxx
  };
  out = FORM["\
#-
#include harmpol.h
#call htables([%2])

* Expr(x) -> int(x,0,1) x^(N-1) Expr(x).
#procedure Mellin(x,N)
  multiply Mel * 1/`x';

  #call hmellin(H,`x',S,`N')
  id sign_(N?) = sign(N);
  .sort:mell1;

  id H(R(?mm),1) = HH(?mm);
  argexplode HH;
  #do i=1,[%2]
    id HH(m1?,...,m`i'?) = Htab`i'(m1,...,m`i');
  #enddo
  id HH(?mm) = H(R(?mm),1);
  .sort:mell2;

  #call Simplify(`N')
#endprocedure

#procedure Simplify(N)
  splitarg (`N'),S;
  repeat;
    id S(R(m1?,?mm),n?pos_,`N')
      = S(R(m1,?mm),n-1,`N')
        + (theta_(m1) + thetap_(-m1) * sign(`N'+n))
        * den(`N'+n)^abs_(m1) * S(R(?mm),n,`N');
    id S(R(m1?,?mm),n?neg_,`N')
      = S(R(m1,?mm),n+1,`N')
        - (theta_(m1) + thetap_(-m1) * sign(`N'+n+1))
        * den(`N'+n+1)^abs_(m1) * S(R(?mm),n+1,`N');
  endrepeat;
  id S(R,n?,N?) = 1;
  id S(R(?mm),0,N?) = S(R(?mm),N);
  .sort:simpl1;

  #call basis(S)
  .sort:simpl2;

  splitarg (`N'),sign;
  id sign(n?int_,`N') = sign_(n) * sign(`N');
  id sign(`N')^2 = 1;
  .sort:simpl3;
#endprocedure

S tmpxxx,[1+tmpxxx],[1-tmpxxx],tmpnnn;

L F = [%1];

#call Mellin(tmpxxx,tmpnnn)

P;
.end", ex, MaxWeight];
  If[out === Null,
    Print["Error: failed to perform the Mellin transform of " <>
          ToString[expr] <> " in FORM"];
    Abort[];
  ];
  out[[1]][[2]] //. {
    S[R[m__], y_] -> HSum[{m}, y],
    den[y_] -> 1/y,
    sign[y_] -> (-1)^y,
    tmpnnn -> n,
    ln2 -> Log[2],
    z2  -> Zeta[2],
    z3  -> Zeta[3],
    z5  -> Zeta[5],
    z7  -> Zeta[7],
    li4half -> PolyLog[4,1/2],
    li5half -> PolyLog[5,1/2],
    li6half -> PolyLog[6,1/2],
    li7half -> PolyLog[7,1/2],
    li8half -> PolyLog[8,1/2],
    s6  -> HPLs6,
    s7a -> HPLs7a,
    s7b -> HPLs7b,
    s8a -> HPLs8a,
    s8b -> HPLs8b,
    s8c -> HPLs8c,
    s8d -> HPLs8d
  }
];

(******************************************************************************)

ExpandQ[e_Times] := MemberQ[e, _Plus|Power[_Plus, p_Integer /; p >= 2], {1}];

ExpandQ[Power[_Plus, p_Integer /; p >= 2]] := True;

ExpandQ[_] := False;

LogExpand[expr_] := expr /. Log[n_] /; IntegerQ[n] && n >= 4 :>
  Plus @@ (#[[2]] Log[#[[1]]] & /@ FactorInteger[n]);

(******************************************************************************)

ClearAll[MTMellinn];

MTMellinn[c_ * f_., x_, n_] := c * MTMellinn[f, x, n] /; FreeQ[c, x] && c =!= 1;

MTMellinn[x_^a_. * f_., x_, n_] := MTMellinn[f, x , n + a];

MTMellinn[e_Plus, x_, n_] := MTMellinn[#, x, n]& /@ e;

MTMellinn[e:_Times|_Power, x_, n_] := MTMellinn[e // Expand, x, n] /;
  ExpandQ[e];

(******************************************************************************)

MTMellinn[DReg[1, x_], x_, n_] := 1;
MTMellinn[DReg[f_HPL, x_], x_, n_] :=
  mtRegTerm[f, x] - Cancel[(n-1) * MTMellinn[f, x, n-1]] // Expand;

mtRegTerm[f_HPL, x_] := HPLLogExtract[f] /. x -> 1 /. HPL[{1}, 1] -> 0;

(******************************************************************************)

ClearAll[MTInverse];

MTInverse[MTMellinn[e_, x_, n_], x_, n_] := e;

MTInverse[c_ * f_., x_, n_] := c * MTInverse[f, x, n] /; FreeQ[c, n] && c =!= 1;

MTInverse[e_Plus, x_, n_] := MTInverse[#, x, n]& /@ e;

MTInverse[e:_Times|_Power, x_, n_] := MTInverse[e // Expand, x, n] /;
  ExpandQ[e];

(******************************************************************************)

MTInverse[e:(n_+a_)^b_. * f_., x_, n_] :=
  x^a * MTInverse[Expand[Apart[e /. n -> n-a, n]], x, n];

MTInverse[e:HSum[l_, n_+a_] *f_., x_, n_] :=
  x^a * MTInverse[Expand[e /. n -> n-a], x, n];

MTInverse[(-1)^(a_+n_) * f_., x_, n_] := (-1)^a * MTInverse[(-1)^n * f, x, n];

(******************************************************************************)

MTPair[m_List] := Module[{ex, ret},
  ex = HPL[m, x];
  ret = HMellin[ex, x, n] // Expand;
  {ex, ret}
];

MTPairs[w_Integer, negative_:True] := Module[{f, ret},
  ret = {};
  If[w == 0,
    ret = Append[ret, {1, 1/n}]
  ,
    If[negative,
      f[{}] := {{1}, {0}, {-1}};
      f[x_] := Flatten[Table[Append[#, i], {i, 1, -1, -1}] & /@ x, 1]
    ,
      f[{}] := {{1}, {0}};
      f[x_] := Flatten[Table[Append[#, i], {i, 1, 0, -1}] & /@ x, 1]
    ];
    (ret = Append[ret, MTPair[#]])& /@ Nest[f, {}, w]
  ];
  ret
];

WriteMTTable[fname_String, MTf_Symbol, t_List] := Module[{str, f},
  If[!FreeQ[t, E],
    Print["Warning: E found in " <> ToString[MTf] <> "[] for " <> fname];
  ];
  If[!FreeQ[t, Log[n_?(#=!=2&)]],
    Print["Warning: Log[N] (N != 2) found in " <> ToString[MTf] <>
          "[] for " <> fname];
  ];
  str = OpenWrite[fname];
  f[e_] := Module[{a},
    a = Switch[MTf,
      MTMellinn, e[[1]] /. x -> x_,
      MTInverse, e[[1]] /. n -> n_
    ];
    WriteString[str, STR[MTf] <> "[" <> STR[a] <> ", x_, n_] = "
                     <> STR[e[[2]]] <> ";\n"];
  ];
  Scan[f, t];
  Close[str];
  Print["MakeTab.m: write " <> ToString[Length[t]] <> " entries to " <> fname];
];

STR[expr_] := expr // InputForm // ToString;

(*
 * NOTE: Simplify[] transforms 2*Log[2] -> Log[4] and sometimes (-1)^(n+a) ->
 * E^(Pi*I*(n+a)) etc.
 *)
SimplifyN[expr_, n_] :=
  Simplify[expr] /. {
    E^(I*Pi*(n+a_.)) /; IntegerQ[a] -> (-1)^n * (-1)^a
  } // LogExpand;

SimplifyN1[expr_, n_] := Module[{ex, nn},
  ex = expr;
  ex = ex /. n -> nn + 1;
  ex = SimplifyN[ex, nn];
  ex = ex /. nn -> n - 1;
  ex
];

(******************************************************************************)

Print["MakeTab.m: weight : " <> ToString[MaxWeight]];

(* Construct a list of pairs {HPL, the Mellin transform of HPL}. *)
t = {};
Do[t = Join[t, MTPairs[i]], {i, 0, MaxWeight}];
t = { #[[1]], SimplifyN[#[[2]], n] } & /@ t;

(* Output the table of Mellin transforms. *)
WriteMTTable[TabFileName, MTMellinn, t];

(* Load the saved table. *)
Get[TabFileName];

(* Output the table of Mellin transforms for regularized derivatives. *)
If[MakeDRegTab,
  tt = {
    DReg[#[[1]], x],
    SimplifyN1[MTMellinn[DReg[#[[1]], x], x, n], n]
  } & /@ t;
  WriteMTTable[DRegTabFileName, MTMellinn, tt];
];

(*
 * Set up the system of the equations for inverse Mellin transforms.
 *
 * NOTE: If the simplification of the equations is imcomplete, it can cause
 *       irrelevant terms in the sysmtem of equations, which actually cancel to
 *       zero, and can lead a wrong result.
 *)
ts = Select[t, #[[1]] === 1 || (Head[#[[1]]] === HPL &&
               (#[[1, 1, 1]] === 1 || #[[1, 1, 1]] === -1))&];
ts = (DReg[#[[1]], x])& /@ ts;
ts = { #, SimplifyN1[MTMellinn[#, x, n], n] }& /@ ts;

eqs1 = #[[1]] == MTInverse[#[[2]], x, n] & /@ t;
eqs2 = #[[1]] == MTInverse[#[[2]], x, n] & /@ ts;

eqs  = Select[Join[eqs1, eqs2], !FreeQ[#, MTInverse] &];
vars = Cases[Variables[#[[2]]& /@ eqs], _MTInverse];

(* Choose the varaibles. *)
vars = Join[
  Cases[vars, MTInverse[1, x, n]],
  Cases[vars, MTInverse[n^_, x, n]],
  Cases[vars, MTInverse[HSum[l_, n], x, n] /;
              (EvenQ[Count[l, i_ /; Negative[i]]])],
  Cases[vars, MTInverse[(-1)^n * HSum[l_, n], x, n] /;
              (OddQ[Count[l, i_ /; Negative[i]]])],
  Cases[vars, MTInverse[n^_. * HSum[l_, n], x, n] /;
              (EvenQ[Count[l, i_ /; Negative[i]]])],
  Cases[vars, MTInverse[(-1)^n * n^_. * HSum[l_, n], x, n] /;
              (OddQ[Count[l, i_ /; Negative[i]]])]
];

(* E^(I*Pi*n) is dangerous! *)
If[!FreeQ[eqs, E],
  Print["Warning: E found in eqs"];
];

(* # of vars must be # of eqs, i.e., 5/2*(3^w-1)+2. *)
Print["MakeTab.m: # of eqs: " <> ToString[Length[eqs]]];
If[Length[eqs] != Length[vars],
  Print["Warning: Length[eqs] != Length[vars]"];
  Print["MakeTab.m: # of vars: " <> ToString[Length[vars]]];
];

(* Solve the system. *)
s = Solve[eqs, vars][[1]];
s = { #[[1]][[1]], SimplifyN[#[[2]], n] } & /@ s;

(* Output the table of inverse Mellin transforms. *)
WriteMTTable[InvTabFileName, MTInverse, s];

(* vim: set ft=mma et ts=8 sts=2 sw=2: *)
